// -*- tab-width: 4; indent-tabs-mode: nil -*-
#ifndef HDNUM_MATRIX_HH
#define HDNUM_MATRIX_HH

#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>

#include "countablearray.hh"
#include "exceptions.hh"
#include "vector.hh"

/** @file
 *  @brief This file implements a generic and dynamic matrix class
 */

namespace hdnum {

  /** @brief A flexible matrix class 

      Matrix is implemented as a handle to a dynamically allocated array. Copy
      and assignement operators have reference semantics. In order to make a true copy 
      the function "copy" can be used.

   */
  template<class T>
  class Matrix {

	typedef CountableArray<T> ArrayType;

  public:

	//! Remember the storage type 
	typedef T value_type;

	/** \brief Reference to an object */
	typedef value_type& reference;

	/** \brief Const reference to an object */
	typedef const value_type& const_reference;

	/** \brief Type used for array indices */
	typedef std::size_t size_type;

	/** \brief Difference type */
	typedef std::ptrdiff_t difference_type;

	//! make empty matrix
	Matrix () : base(), m(0), n(0), stride(0), data(0)
	{}

	//! make _m x _n matrix uninitialized
	Matrix (size_type _m, size_type _n) 
	  : base(new ArrayType(_m*_n)), m(_m), n(_n), stride(_n)
	{
	  data = &((*base)[0]);
	}

	//! make _m x _n matrix initialized
	Matrix (size_type _m, size_type _n, const T& _t) 
	  : base(new ArrayType(_m*_n,_t)), m(_m), n(_n), stride(_n)
	{
	  data = &((*base)[0]);
	}

    //! copy constructor with reference semantics
	Matrix (const Matrix& A) 
      : base(A.base), m(A.m), n(A.n), stride(A.stride), data(A.data)
    {}

    //! assignment operator with reference semantics
	Matrix& operator= (const Matrix& A)
    {
      base = A.base;
      m = A.m;
      n = A.n;
      stride = A.stride;
      data = A.data;
      return *this;
    }

    //! assignment from scalar
	Matrix& operator= (const T& t)
    {
      for (size_type i=0; i<rowsize(); i++)
        for (size_type j=0; j<colsize(); j++)
          (*this)[i][j] = t;
      return *this;
    }

    //! submatrix extraction
    Matrix<T> sub (size_type i, size_type j, size_type rows, size_type cols)
    {
      Matrix<T> A;
      A.base = base;
      A.stride = stride;
      A.m = rows;
      A.n = cols;
      A.data = data+i*stride+j;
      return A;
    }

	//! Component access
	T* operator[] (size_type i)
	{
	  return data+i*stride;
	}

	//! Component access
	const T* operator[] (size_type i) const
	{
	  return data+i*stride;
	}

     /*! 
      \brief Addition assignment 

      Implements A += B matrix  addition
 
      \param[in] B another Matrix
    */
    Matrix& operator+= (const Matrix& B)
    {
      for (size_type i=0; i<m; ++i) 
        for (size_type j=0; j<n; ++j) 
        (*this)[i][j] += B[i][j];
      return *this;
    }

     /*! 
      \brief Subtraction assignment

      Implements A -= B matrix subtraction
 
      \param[in] B another matrix
    */
    Matrix& operator-= (const Matrix& B)
    {
      for (size_type i=0; i<m; ++i) 
        for (size_type j=0; j<n; ++j) 
        (*this)[i][j] -= B[i][j];
      return *this;
    }

     /*! 
      \brief Scalar multiplication assignment

      Implements A *= s where s is a scalar
 
      \param[in] s scalar value to multiply with
    */
    //! vector scalar multiplication
    Matrix& operator*= (const T& s)
    {
      for (size_type i=0; i<m; ++i) 
        for (size_type j=0; j<n; ++j) 
          (*this)[i][j] *= s;
      return *this;
    }

     /*! 
      \brief Scalar division assignment

      Implements A /= s where s is a scalar
 
      \param[in] s scalar value to multiply with
    */
    //! vector scalar multiplication
    Matrix& operator/= (const T& s)
    {
      for (size_type i=0; i<m; ++i) 
        for (size_type j=0; j<n; ++j) 
          (*this)[i][j] /= s;
      return *this;
    }

    /*! 
      \brief Scaled update of a Matrix

      Implements A += s*B where s is a scalar and B a matrix
 
      \param[in] s scalar value to multiply with
      \param[in] B another matrix
    */
   void update (const T& s, const Matrix& B)
    {
      for (size_type i=0; i<m; ++i) 
        for (size_type j=0; j<n; ++j) 
          (*this)[i][j] += s*B[i][j];
    }


    //! matrix vector product y = A*x
    template<class V>
    void mv (Vector<V>& y, const Vector<V>& x) const
    {
      if (this->rowsize()!=y.size()) 
        HDNUM_ERROR("mv: size of A and y do not match");
      if (this->colsize()!=x.size()) 
        HDNUM_ERROR("mv: size of A and x do not match");
      for (size_type i=0; i<rowsize(); i++)
        {
          y[i] = 0;
          for (size_type j=0; j<colsize(); j++)
            y[i] += (*this)[i][j]*x[j];
        }
    }
    
    //! update matrix vector product y += A*x
    template<class V>
    void umv (Vector<V>& y, const Vector<V>& x) const
    {
      if (this->rowsize()!=y.size()) 
        HDNUM_ERROR("mv: size of A and y do not match");
      if (this->colsize()!=x.size()) 
        HDNUM_ERROR("mv: size of A and x do not match");
      for (size_type i=0; i<rowsize(); i++)
        {
          for (size_type j=0; j<colsize(); j++)
            y[i] += (*this)[i][j]*x[j];
        }
    }

    //! update matrix vector product y += A*x
    template<class V>
    void umv (Vector<V>& y, const V& s, const Vector<V>& x) const
    {
      if (this->rowsize()!=y.size()) 
        HDNUM_ERROR("mv: size of A and y do not match");
      if (this->colsize()!=x.size()) 
        HDNUM_ERROR("mv: size of A and x do not match");
      for (size_type i=0; i<rowsize(); i++)
        {
          for (size_type j=0; j<colsize(); j++)
            y[i] += s*(*this)[i][j]*x[j];
        }
    }

    //! matrix matrix product C = A*B
    void mm (const Matrix<T>& A, const Matrix<T>& B)
    {
      if (this->rowsize()!=A.rowsize()) 
        HDNUM_ERROR("mm: size incompatible");
      if (this->colsize()!=B.colsize()) 
        HDNUM_ERROR("mm: size incompatible");
      if (A.colsize()!=B.rowsize()) 
        HDNUM_ERROR("mm: size incompatible");

      for (size_type i=0; i<rowsize(); i++)
        for (size_type j=0; j<colsize(); j++)
          {
            (*this)[i][j] = 0;
            for (size_type k=0; k<A.colsize(); k++)
              (*this)[i][j] += A[i][k]*B[k][j];
          }
    }
    
    //! matrix matrix product C += A*B
    void umm (const Matrix<T>& A, const Matrix<T>& B)
    {
      if (this->rowsize()!=A.rowsize()) 
        HDNUM_ERROR("mm: size incompatible");
      if (this->colsize()!=B.colsize()) 
        HDNUM_ERROR("mm: size incompatible");
      if (A.colsize()!=B.rowsize()) 
        HDNUM_ERROR("mm: size incompatible");

      for (size_type i=0; i<rowsize(); i++)
        for (size_type j=0; j<colsize(); j++)
          for (size_type k=0; k<A.colsize(); k++)
            (*this)[i][j] += A[i][k]*B[k][j];
    }
    
    //! set column: make x the k'th column of A
    void sc (const Vector<T>& x, size_type k)
    {
      if (this->rowsize()!=x.size()) 
        HDNUM_ERROR("cc: size incompatible");

      for (size_type i=0; i<rowsize(); i++)
        (*this)[i][k] = x[i];
    }

    //! set row: make x the k'th row of A
    void sr (const Vector<T>& x, size_type k)
    {
      if (this->colsize()!=x.size()) 
        HDNUM_ERROR("cc: size incompatible");

      for (size_type i=0; i<colsize(); i++)
        (*this)[k][i] = x[i];
    }

    //! compute row sum norm
    T norm_infty () const
    {
      T norm(0.0);
      for (size_type i=0; i<rowsize(); i++)
        {
          T sum(0.0);
          for (size_type j=0; j<colsize(); j++)
            sum += myabs((*this)[i][j]);
          if (sum>norm) norm = sum;
        }
      return norm;
    }

    //! compute column sum norm
    T norm_1 () const
    {
      T norm(0.0);
      for (size_type j=0; j<colsize(); j++)
        {
          T sum(0.0);
          for (size_type i=0; i<rowsize(); i++)
            sum += myabs((*this)[i][j]);
          if (sum>norm) norm = sum;
        }
      return norm;
    }

	//! get number of rows
	size_type rowsize () const
	{
	  return m;
	}

	//! get number of columns
	size_type colsize () const
	{
	  return n;
	}

    //! get index field width for pretty-printing
    int iwidth () const
    {
      return iwidth_;
    }

    //! get data field width for pretty-printing
    int width () const
    {
      return width_;
    }

    //! get data precision for pretty-printing
    int precision () const
    {
      return precision_;
    }

    //! set index field width for pretty-printing
    void iwidth (int i) const
    {
      iwidth_=i;
    }

    //! set data field width for pretty-printing
    void width (int i) const
    {
      width_=i;
    }

    //! set data precision for pretty-printing
    void precision (int i) const
    {
      precision_=i;
    }

  private:
    T myabs (T x) const
    {
      if (x>=T(0)) return x; else return -x; 
    }

	CP<ArrayType> base; // array referenced by this vector
	size_type m,n;      // dimensions of the matrix (m x n : m rows, n columns)
	size_type stride;   // original length of a row
	T* data;            // pointer to first element used in base
    static int iwidth_, width_, precision_;
  };

  template<class T>
  int Matrix<T>::iwidth_ = 3;
  template<class T>
  int Matrix<T>::width_ = 10;
  template<class T>
  int Matrix<T>::precision_ = 2;

  //! Output operator for Matrix
  template <class T>
  inline std::ostream& operator<< (std::ostream& s, const Matrix<T>& A)
  {
	s << " " << std::setw(A.iwidth()) << " " << "  ";
	for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
	  s << std::setw(A.width()) << j << " ";
	s << std::endl;
	for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  {
		s << " " << std::setw(A.iwidth()) << i << "  ";
		for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
		  s << std::setw(A.width()) << std::scientific << std::showpoint 
			<< std::setprecision(A.precision()) << A[i][j] << " "; 
		s << std::endl;
	  }
    return s;
  }

  //! make a symmetric and positive definite matrix
  template<class T>
  inline void fill (Matrix<T> A, const T& t)
  {
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
        A[i][j] = t;
  }

  //! make a symmetric and positive definite matrix
  template<class T>
  inline void zero (Matrix<T> A)
  {
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
        A[i][j] = T(0);
  }

  //! make identity matrix
  template<class T>
  inline void identity (Matrix<T> A)
  {
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
		if (i==j)
		  A[i][i] = T(1);
		else
		  A[i][j] = T(0);
  }

  //! make a symmetric and positive definite matrix
  template<class T>
  inline void spd (Matrix<T> A)
  {
	if (A.rowsize()!=A.colsize() || A.rowsize()==0) 
	  HDNUM_ERROR("need square and nonempty matrix");
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
		if (i==j)
		  A[i][i] = T(4.0);
		else
		  A[i][j] = - T(1.0)/((i-j)*(i-j));
  }

  //! make a vandermonde matrix
  template<class T>
  inline void vandermonde (Matrix<T> A, const Vector<T> x)
  {
	if (A.rowsize()!=A.colsize() || A.rowsize()==0) 
	  HDNUM_ERROR("need square and nonempty matrix");
	if (A.rowsize()!=x.size()) 
	  HDNUM_ERROR("need A and x of same size");
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
      {
        T p(1.0);
        for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
          {
            A[i][j] = p;
            p *= x[i];
          }
      }
  }

  //! make a true copy of a matrix
  template<class T>
  inline Matrix<T> copy (const Matrix<T> A)
  {
    Matrix<T> B(A.rowsize(),A.colsize());
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
	  for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
        B[i][j] = A[i][j];
    return B;
  }

  //! gnuplot output for matrix
  template<class T>
  inline void gnuplot (const std::string& fname, const Matrix<T> A)
  {
    std::fstream f(fname.c_str(),std::ios::out); 
    for (typename Matrix<T>::size_type i=0; i<A.rowsize(); ++i)
      {
        for (typename Matrix<T>::size_type j=0; j<A.colsize(); ++j)
          f << std::setw(A.width()) << std::scientific << std::showpoint 
            << std::setprecision(A.precision()) << A[i][j] << std::endl;
        f << std::endl;
      }
    f.close();
  }


} // namespace hdnum

#endif
